function [bits_out] = RLF_decode(bits_in, block, blocks, error_map, key)
%
% bits_out = RLF_decode(bits_in, block, blocks, error_map, key)
%
% Implementation of the random linear fountain decoder. Uses Gaussian
% elimination for solving the problem.
% 
% Input Parameters:
%
%   - bits_in - input bitstream, vector of 0 and 1
%
%   - block   - symbol length
%
%   - blocks  - the number of original message symbols
%  
%   - error_map - the tampering map, vector of 0 and 1, tampered as 1-s
%
%   - key - integer, the seed for the PRNG
%
% Output:
% 
% bits_out - decoded bitstream, vector of 0 and 1
%
% ------------------------------------------------------------------------
%
% Copyright (C) Paweł Korus
% All rights reserved.
%
% Contact: Paweł Korus, pkorus@agh.edu.pl
%
% The author is with the Department of Telecommunications of AGH University
% of Science and Technology, Kraków, Poland.
%
%----------------------------------------------------------------------
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%

    blk = length(bits_in)/block;
    D = zeros(1, blk, 'uint16');
    G = zeros(blk, blocks, 'uint16');
    
    rand('twister', key);
    
    for d = 1:blk
        ro = randint(1, blocks, 2); 
        G(d,:) = ro;
        if nargin > 3
            if error_map(d) == true
                G(d,:) = 0;
            end
        end
    end
    
    if rank(double(G)) < blocks
        throw( MException('rlc:insufficientData', 'Not enought input symbols'));
    end

    % Append symbols to G
    D = zeros(length(bits_in)/block, blocks+block);
    for bi=1:length(bits_in)/block
        D(bi,:) = [G(bi,:) bits_in(1+(bi-1)*block:bi*block)];
    end
    
    G = D;
    
    % Step I
    for i = 1:blocks
        % find rows with ones on i-th position
        rows = i-1+find(G(i:end,i));
        if length(rows) == 0
            throw( MException('rlc:insufficientData', 'Not enought input symbols'));
        end
        % select the one with the least ones
        sums = sum(G');
        relevant_sums = sums(rows);
        [x, row_ind] = min(relevant_sums);
        row_ind = rows(row_ind);
        tr = G(row_ind,:);
        % remove the dependency on row_ind form other rows
        for b = rows'
            if b ~= row_ind
                G(b,:) = mod(G(b,:) + tr,2);
            end
        end
        % move the row as the first one
        G(row_ind,:) = G(i,:);
        G(i,:) = tr;
    end
    
    % Step II - back-substitution
        for i = 1:blocks
        % find rows with ones on i-th position
        rows = find(G(1:blocks-i+1,blocks-i+1));
        row_ind = rows(end);
        tr = G(row_ind,:);
        % remove the dependency on row_ind form other rows
        for b = rows'
            if b ~= row_ind
                G(b,:) = mod(G(b,:) + tr,2);
            end
        end
%         % move the row as the first one
%         G(row_ind,:) = G(i,:);
%         G(i,:) = tr;
    end
    
    
    bits_out = G(1:blocks,blocks+1:end);
    bits_out = bits_out';
    bits_out = bits_out(:)';
%     
%     
% %     G = gf(G);
%     [R, lincols] = rref(G'); 
%     G = gf(G(lincols,:)); 
%     if size(G,1) ~= size(G,2)
%         throw( MException('rlc:insufficientData', 'Not enought input symbols'));
%     end
%     Gi = abs(inv(G));
%     bits_in2 = zeros(1, block*blocks);
%     for b = 1:blocks
%         bi = lincols(b);
%         bits_in2(1+(b-1)*block:b*block) = bits_in(1+(bi-1)*block:bi*block);
%     end
%     
%     bits_out = zeros(1, blocks * block);
%     
%     for i = 1:blocks
%         ro = Gi(i,:);
%         P = find(ro);
%         B = zeros(numel(P), block);
%         for b = 1:numel(P)
%             bi = P(b);
%             B(b, :) = bits_in2(1+(bi-1)*block:bi*block);
%         end
%         if numel(P) == 1
%             bits_out(1+(i-1)*block:i*block) = mod(B, 2);
%         else
%             bits_out(1+(i-1)*block:i*block) = mod(sum(B), 2);
%         end
%     end    
end