function [bits_out] = RLF_self_decode(bits_in, block, blocks, error_map, key, bitstream_ref, reconstruction_map)
%
% bits_out = RLF_decode(bits_in, block, blocks, error_map, key, 
%                       bitstream_ref, reconstruction_map)
%
% Implementation of the self-recovery decoder. Simplifies the content
% reconstruction problem by removing the dependencies on known symbols and
% by decoding only the necessary ones. After simplification, uses Gaussian
% elimination for solving the remaining problem.
% 
% Input Parameters:
%
%   - bits_in - input bitstream, vector of 0 and 1
%
%   - block   - symbol length
%
%   - blocks  - the number of original message symbols
%  
%   - error_map - the tampering map, vector of 0 and 1, tampered as 1-s
%
%   - key - integer, the seed for the PRNG
%
%   - bitstream_ref - input bitstream with valid symbols regenerated from
%                     the remaining authentic image fragments, same length
%                     as 'bits_in'
%
%   - reconstruction_map - vector of 0 and 1, marks the original symbols
%                          which actually need to be decoded
%
% Output:
% 
% bits_out - decoded bitstream, vector of 0 and 1
%
% ------------------------------------------------------------------------
%
% Copyright (C) Paweł Korus
% All rights reserved.
%
% Contact: Paweł Korus, pkorus@agh.edu.pl
%
% The author is with the Department of Telecommunications of AGH University
% of Science and Technology, Kraków, Poland.
%
%----------------------------------------------------------------------
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%

    % Pad with 0's
    if mod(length(bitstream_ref), block) ~= 0
        pad = zeros(1, block - mod(length(bitstream_ref), block));
        bitstream_ref = [bitstream_ref pad];
    end

    % Initialize variables
    blk = length(bits_in)/block;
    D = zeros(1, blk, 'uint8');
    G = zeros(blk, blocks, 'uint8');
    
    % re-create the G matrix from the encoder
    rand('twister', key);
    
    for d = 1:blk
        ro = randint(1, blocks, 2);
        G(d,:) = ro;
        if nargin > 3
            if error_map(d) == true
                G(d,:) = 0;
            end
        end
    end
    
%     figure(2); spy(G); title ('After removing erased symbols');
    
    % remove connections to known symbols
    for d = 1:blk
        ro = G(d,:);
        if sum(ro) == 0
            continue;
        end
        % find symbols which are used and are known (authentic)
        for bi = find(ro .* (1-reconstruction_map))
            bits_in(1+(d-1)*block:d*block) = mod(bits_in(1+(d-1)*block:d*block) + bitstream_ref(1+(bi-1)*block:bi*block),2);
        end
        % remove the dependency
        ro = mod(ro .* reconstruction_map,2);
        % re-assign
        G(d,:) = ro;
    end
    
    % Convert input bit-streams to a more convenient form
    bits_in_m = zeros(length(bits_in)/block, block);
    
    for d = 1:blk
        bits_in_m(d,:) = bits_in(1+(d-1)*block:d*block);
    end
    
%     figure(3); spy(G); title ('Before Gaussian Elimination');
    
    if rank(double(G)) < sum(reconstruction_map)
        throw( MException('rlc:insufficientData', 'Not enought input symbols'));
    end

    % Cut the part of G which is useful, i.e. not erased and 
    % need to be recovered
    sx = sum(G(:,:),1); sy = sum(G(:,:),2); 
    G = G(find(sy), find(sx));
    blocks_n = size(G,2);
    
    % Cut the irrelevant symbols
    bits_in_m = bits_in_m(find(sy),:);
    
    % Append symbols to G
    D = zeros(size(bits_in_m,1), blocks_n+block);
    for bi=1:size(bits_in_m,1)
        D(bi,:) = [G(bi,:) bits_in_m(bi,:)];
    end
    
    G = D;
    
    % Step I
    for i = 1:blocks_n        
        % find rows with ones on i-th position
        rows = i-1+find(G(i:end,i));
        if isempty(rows)
            throw( MException('rlc:algorithmError', 'Cannot find informative rows!'));
        end
        % select the one with the least ones
        sums = sum(G(:,1:blocks_n),2);
        relevant_sums = sums(rows);
        [x, row_ind] = min(relevant_sums);
        row_ind = rows(row_ind);
        tr = G(row_ind,:);
        % remove the dependency on row_ind form other rows
        for b = rows'
            if b ~= row_ind
                G(b,:) = mod(G(b,:) + tr,2);
            end
        end
        % move the row as the first one
        G(row_ind,:) = G(i,:);
        G(i,:) = tr;
    end
    
%     figure(4); spy(G(:,1:blocks_n)); title ('Before back-subst');
    
    % Step II - back-substitution
    for i = 1:blocks_n
        % find rows with ones on i-th position
        rows = find(G(1:blocks_n-i+1,blocks_n-i+1));
        if isempty(rows)
            throw( MException('rlc:insufficientData', 'Not enought input symbols'));
        end
        row_ind = rows(end);
        tr = G(row_ind,:);
        % remove the dependency on row_ind form other rows
        for b = rows'
            if b ~= row_ind
                G(b,:) = mod(G(b,:) + tr,2);
            end
        end
    end
    
%     figure(5); spy(G(:,1:blocks_n)); title ('Final');
    
    % Restore original bits order
    bits_out = zeros(blocks, block);
    gi = 1;
    for d = find(sx)
        bits_out(d,:) = G(gi, blocks_n+1:end);
        gi = gi + 1;
    end
    
    bits_out = bits_out';
    bits_out = bits_out(:)';
    bits_out = uint16(bits_out);
end