function [img_low_quality, psnr] = get_reference_high(img, q_bits, dr_factor)
%
% [img_low_quality, psnr] = get_reference(img, q_bits)
%
% Gets the reference image for a specified number of reference bits per block
%
% Input parameters:
%
%   - img - uint8 matrix representing a single channel grayscale image with
%           the size of multiples of 8
%
%   - q_bits - number of bits of rec. reference per image block (between 1
%
% Output:
% 
%   - img_low_quality - the resulting reference image
%   - psnr - the peak signal to noise ratio
% ------------------------------------------------------------------------
%
% Copyright (C) Paweł Korus
% All rights reserved.
%
% Contact: Paweł Korus, pkorus@agh.edu.pl
%
% The author is with the Department of Telecommunications of AGH University
% of Science and Technology, Kraków, Poland.
%
%----------------------------------------------------------------------
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%

    [H, W] = size(img);
            
    % PREPARE A LOW-DETAIL VERSION OF THE IMAGE.
    block = 8;

    % Limit the dynamic range of the image
    img_low_dr = double(floor(double(img)/dr_factor));

    Q = [8 6 2 1 1 1 1 1
         6 2 1 1 1 1 1 1
         2 1 1 1 1 1 1 1
         1 1 1 1 1 1 1 1
         1 1 1 1 1 1 1 1
         1 1 1 1 1 1 1 1
         1 1 1 1 1 1 1 1
         1 1 1 1 1 1 1 1];
    Q = max(ones(size(Q)),Q/dr_factor);
    
    T = dctmtx(8);
    f = @(x) (T*x*T')./Q;
    fi = @(x) T'*(x.*Q)*T;

    S_b = blkproc(img_low_dr, block*[1, 1], f);

    % Matrix P contains the number of bits that can be used to
    % represent each coefficient in a block, for maximum quality
    P = get_reference_alloc_highq(q_bits);

    for bx = 1:W/block
        for by = 1:H/block
            B = S_b(1+(by-1)*block:(by)*block, 1+(bx-1)*block:(bx)*block);
            B = round(B);            

            % Matrix X is the maxmum absolute value that can be encoded for
            % each coefficient, as defined by P
            M = pow2(P-1)-1;
            M(1,1) = pow2(P(1,1))-1;

            % Saturate the values that exceed M
            B(B > M) = M(B > M);
            B(B < -M) = -(M(B < -M));

            % Zero the coefficients that have no anticipated bits
            B(P == 0) = 0;

            S_b(1+(by-1)*block:(by)*block, 1+(bx-1)*block:(bx)*block) = B;
        end
    end

    img_low_quality = dr_factor*uint8(blkproc(S_b, block*[1, 1], fi));
    img_low_quality = bitset(img_low_quality, log2(dr_factor), 1);
    psnr = calc_psnr(img, img_low_quality);
end