function [imgw, r] = c_encode(img, options, QProfiles)
%
% Copyright (C) 2012 Paweł Korus <pkorus@agh.edu.pl>
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%

    [W, H] = size(img);
    r = struct();
    
    % Read options
    LSBS = options.LSBS;
    bit_weights = zeros(1,LSBS, 'uint16');
    for i = 1:LSBS
        bit_weights(i) = 2^(i-1);
    end
    dr_factor = pow2(options.LSBS);
    BLOCK = options.BLOCK;
    
    % Symbol length for the QDF
    SQ_LENGTH = options.SQ_LENGTH;

    % Symbol length for the reconstruction reference
    S_LENGTH = options.BLOCK*options.BLOCK*options.LSBS-options.H_LENGTH - SQ_LENGTH;
    
    KEY = options.KEY;
                
    % PREPARE A LOW-DETAIL VERSION OF THE IMAGE.

    % Limit the dynamic range of the image
    img_low_dr = double(floor(double(img)/dr_factor));
    
    % Initialize quality profiles
    P_profiles = QProfiles.P_profiles;
    LM_quantas = QProfiles.LM_quantas;
        
    if isfield(options, 'QD')
        QD = options.QD;
    else
        QD = map_profiles(img, QProfiles);
    end
    
    
    % Bits needed for the quality descriptor
    QD_capacity = numel(QD)*(options.qdf_resolution);
    
    if options.verbose == true
        fprintf('Image: %d x %d px -> %d blocks\n', W, H, W*H/BLOCK/BLOCK);
        fprintf('Using %d LSBs -> %d bits of payload (%d b / block)\n', LSBS, W*H*LSBS, BLOCK*BLOCK*LSBS);
        fprintf('  - %3d b for hash\n', options.H_LENGTH);
        fprintf('  - %3d b for descriptor symbol\n', SQ_LENGTH);
        fprintf('  - %3d b for reference symbol\n', S_LENGTH);
        fprintf('QD with %d distinct levels: [', length(unique(QD(:))));
        fprintf(' %d', unique(QD(:)));
        fprintf(']\n');
        fprintf('QD coded on %d bits (%d per block)\n', QD_capacity, options.qdf_resolution);
    end
            
    % Transform the low-dynamic range image into spectral domain
    T = dctmtx(8);
    f = @(x) (T*x*T');
    fi = @(x) T'*(x)*T;

    S_b = blkproc(img_low_dr, BLOCK*[1, 1], f);

    % Get a mapping for the zig-zag order

    indices = zigzag(BLOCK);

    % bits needed for the reconstruction stream and for the qdf
    qdf_bitstream = zeros(1,QD_capacity, 'uint16');
    
    Ps = zeros(H/BLOCK, W/BLOCK);

    for bx = 1:W/BLOCK
        for by = 1:BLOCK
            Ps(by,bx) = P_profiles(QD(by, bx));
        end
    end
    
    bitstream = zeros(1,sum(sum(Ps)), 'uint16');

    % GENERATE THE RECONSTRUCTION REFERENCE

    % pointer to the qdf data
    bqi = 1;
    % pointer to the reconstruction stream data
    bi = 1;

    for bx = 1:W/BLOCK
        for by = 1:H/BLOCK
            B = S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
            % Based on QDF for this block, get quantization matrix
            c = QD(by, bx);
            P = P_profiles(:,:,c);
            
            % Quantize the block
            [B, I] = quantize_block(B, P, 1, LM_quantas{c});
            
            % Encode the QDF
            cbits = dec2binvec(c-1, options.qdf_resolution);
            if length(cbits) > options.qdf_resolution
                throw(MException('encoder:QDError', 'QD value out of scope'));
            end
            qdf_bitstream(bqi:bqi+length(cbits)-1) = cbits;
            bqi = bqi+length(cbits);
            
            if sum(sum(P)) == 0
                continue
            end

            % Encode the coefficients into a bitstream
            V = I(indices);
            Q = P(indices);
            Q(Q == 0) = [];
            bits = false(1, sum(sum(P)));
            last_bits_index = 0;

            for i = 1:length(Q)
                v = V(i);
                q = Q(i);
                % Encode the index of a code-book entry
                bits(last_bits_index+1:last_bits_index+q) = dec2binvec(v-1, q);
                last_bits_index = last_bits_index + q;    
            end

            bitstream(bi:bi+length(bits)-1) = bits;
            bi = bi+length(bits);

            S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
        end
    end
    
    r.bitstream = bitstream;

    % Get a spatial-domain representation of the reference image
    img_low_quality = blkproc(S_b, BLOCK*[1, 1], fi);
    img_low_quality = uint8(round(dr_factor*img_low_quality+pow2(-1+log2(dr_factor))*ones(size(img))));
    r.img_low_quality = img_low_quality;
    
    % Calculate reconstruction reference quality
    r.ref_psnr = calc_psnr(img, img_low_quality);
        
    % ENCODE THE RECONSTRUCTION STREAM WITH THE RLF CODE
    capacity = S_LENGTH*W*H/BLOCK/BLOCK;
    symbols_needed = capacity/S_LENGTH;
    if options.use_mex == true
        bitstream_enc = c_rlf_encode(bitstream, S_LENGTH, symbols_needed, KEY);
        qdf_bitstream_enc = c_rlf_encode(qdf_bitstream, SQ_LENGTH, symbols_needed, KEY);
    else
        bitstream_enc = RLF_encode(bitstream, S_LENGTH, symbols_needed, KEY);
        qdf_bitstream_enc = RLF_encode(qdf_bitstream, SQ_LENGTH, symbols_needed, KEY);
    end
    r.bitstream_enc = bitstream_enc;
    
    if options.verbose == true
        fprintf('Reference stream %d bits (%d symbols)\n', length(bitstream), ceil(length(bitstream)/S_LENGTH));
        fprintf('Reference stream uses %.2f%% of capacity\n', 100*length(bitstream)/capacity);
    end
    
    % EMBED THE WATERMARK

    imgw = img;

    blocks_no = ceil(length(bitstream)/S_LENGTH);
    if options.verbose == true
        fprintf('Input symbols: %d\n', blocks_no);
        fprintf('Output symbols: %d\n', symbols_needed);
        fprintf('Payload / Capacity Ratio: %.3f\n', blocks_no/symbols_needed);
    end

    for bx = 1:W/BLOCK
        for by = 1:H/BLOCK
            % Determine block ID in scale 1, ..., WH/b/b
            block_id = by + (bx-1)*H/BLOCK - 1 + 1;

            % Get block data for original and low-rate versions
            B = img(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
            B_a = uint8(img_low_dr(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK));

            % Take:
            % 1) block ID,
            % 2) low quality block data (for reconstruction),
            % 3) low dynamic range block data (for tampering detection).
            % ...
            bits_b = uint16(block_id);
            bits_d = bitstream_enc(1+(block_id-1)*S_LENGTH:block_id*S_LENGTH);
            bits_q = qdf_bitstream_enc(1+(block_id-1)*SQ_LENGTH:block_id*SQ_LENGTH);
            bits_a = uint16(B_a(:)'); 

            % and supply it into a hashing function
            h = md5([bits_b bits_a bits_d bits_q uint16(dec2binvec(options.KEY, 32))]);

            % Since classic MD5 yields 128 bit output, make a short vesion (32
            % bit) by summming up binary symbol data
            h_short = zeros(1, 32);
            for i=1:length(h)
                h_short(i) = mod(uint8(h(i)), 2);
            end

            % Replace 3 LSBs with the hash value and low quality block data
            bits = [h_short bits_q bits_d];

            for i=1:numel(B_a)
                B(i) = dr_factor*B_a(i) + uint8(sum(bits(1+(i-1)*LSBS:i*LSBS).*bit_weights));
            end

            imgw(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
        end
    end

    % Calculate the watermarked image PSNR
    r.psnr = calc_psnr(img, imgw);
    r.QD = QD;
    r.S_LENGTH = S_LENGTH;
    r.SQ_LENGTH = SQ_LENGTH;
    
    if options.verbose == true
        fprintf('Reference quality: %.2f dB\n', r.ref_psnr);
        fprintf('Watermarked quality: %.2f dB\n', r.psnr);
    end
end