function [img_fin, r] = c_ref_decode(imgt, options, QProfiles, profile_id)
%
% Copyright (C) 2012 Paweł Korus <pkorus@agh.edu.pl>
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%

    [W, H] = size(imgt);
    r = struct();
    
    % Read options
    method = options.decoder;
    LSBS = options.LSBS;
    MSBS = 8 - LSBS;
    dr_factor = pow2(options.LSBS);
    BLOCK = options.BLOCK;
    S_LENGTH = options.BLOCK*options.BLOCK*options.LSBS-options.H_LENGTH;
    RANDOMIZE_LSBS = options.RANDOMIZE_LSBS;
    KEY = options.KEY;
            
    % PREPARE A LOW-DETAIL VERSION OF THE IMAGE.

    % Limit the dynamic range of the image
    img_low_dr = double(floor(double(imgt)/dr_factor));

    % Initialize quality profiles
    P_profiles = QProfiles.P_profiles;
    LM_quantas = QProfiles.LM_quantas;
        
    % TODO Get the suggested quality descriptor
%     QD_suggested = get_suggested_qdf(img_low_dr);
    
    % TODO Combine the suggested quality descriptor
    QD = 2*ones([H, W]/BLOCK);
    
    % TODO remember to consider options.use_old_reference
    
    % Bits needed for the quality descriptor
    QD_capacity = numel(QD)*(options.qdf_resolution);
    
    % TODO ----------------------------------------------------------------
    
    T = dctmtx(8);
    f = @(x) (T*x*T');
    fi = @(x) T'*(x)*T;

    % Get a mapping for the zig-zag order

    indices = zigzag(BLOCK);

    imgt2 = imgt;

    bits_recovered = zeros(1, W/BLOCK*H/BLOCK*S_LENGTH, 'uint16');
    error_map = zeros(1, W/BLOCK*H/BLOCK, 'uint8');

    for bx = 1:W/BLOCK
        for by = 1:H/BLOCK

            % Generate block id in range 1, ..., WH/b/b
            block_id = by + (bx-1)*H/BLOCK - 1 + 1;
            
            % Get block data for original and low-rate versions
            B = imgt(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
            B_a = uint8(img_low_dr(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK));
            
            % Read the embedded bits from the image
            bits = zeros(1, LSBS*BLOCK*BLOCK);
            for i=1:numel(B_a)
                bits(1+(i-1)*LSBS:i*LSBS) = dec2binvec(B(i), LSBS);
            end

            % Read the embedded hash value
            h_short_read = bits(1:32);

            % Regenerate the bit sequence used to feed the hashing function            
            bits_b = uint16(block_id);
            bits_d = bits((1+32):end);
            bits_a = uint16(B_a(:)'); 

            % Perform hashing
            h = md5([bits_b bits_a bits_d uint16(dec2binvec(options.KEY, 32))]);
            h_short = zeros(1, 32);
            for i=1:length(h)
                h_short(i) = mod(uint8(h(i)), 2);
            end

            % If hashes do not match, the block has been tampered with.
            % In such case:
            % - mark the block as erased (model of channel with erasures)
            % - set the bits recovered from this block to 0
            if ~all(h_short == h_short_read)
                imgt2(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = 1.5*B;
                bits_recovered(1+(block_id-1)*S_LENGTH:block_id*S_LENGTH) = 0;
                error_map(block_id) = true;
            else
                bits_recovered(1+(block_id-1)*S_LENGTH:block_id*S_LENGTH) = bits_d;
            end
        end
    end

    if options.verbose == true
        fprintf('Tampered blocks: %d (%.2f %%)\n', sum(sum(error_map)), 100*sum(sum(error_map))/(W/BLOCK*H/BLOCK));
    end
    
    % Pre-initialize helper variables
    r.rec_psnr = 0;
    ERR = 0;
    img_fin = uint8(zeros([H W]));
    r.reconstruction_demand  = 0;
    P = P_profiles(:,:,profile_id);
    
    % If there are no errors in the identification of tampered fragments
    if sum(sum(error_map)) > 0 && sum(sum(error_map)) < numel(error_map)
        
        blocks_no = ceil(sum(sum(P))*W*H/BLOCK/BLOCK/S_LENGTH);
        r.K = blocks_no;
%         blocks_no = ceil(H/options.BLOCK*W/options.BLOCK*lambda);
        
        % OBTAIN THE RECONSTRUCTION REFERENCES OF AUTHENTIC BLOCKS
        % & THE RECONSTRUCTION MAP - WHICH INPUT SYMBOLS ARE NEEDED
        reconstruction_map = zeros(1, blocks_no, 'uint8');

        % pointer to the reconstruction stream data
        bi = 1;
        bitstream_ref = zeros(1,sum(sum(P))*W*H/BLOCK/BLOCK, 'uint16');

        S_b = blkproc(img_low_dr, BLOCK*[1, 1], f);
    
        for bx = 1:W/BLOCK
            for by = 1:H/BLOCK
                block_id = by + (bx-1)*H/BLOCK - 1 + 1;
                B = S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
        
                % Quantize the block
                [B, I] = quantize_block(B, P, 1, LM_quantas{profile_id});

                % Encode the coefficients into a bitstream
                V = I(indices);
                Q = P(indices);
                Q(Q == 0) = [];
                bits = false(1, sum(sum(P)));
                last_bits_index = 0;

                if error_map(block_id) == false
                    for i = 1:length(Q)
                        v = V(i);
                        q = Q(i);
                        bits(last_bits_index+1:last_bits_index+q) = dec2binvec(v-1, q);
                        last_bits_index = last_bits_index + q;    
                    end
                else
                    f_block_id = 1+floor((bi-1)/ S_LENGTH);
                    l_block_id = 1+floor((bi+length(bits)-2)/S_LENGTH);
                    for m_block_id = f_block_id:l_block_id
                        reconstruction_map(m_block_id) = 1;
                    end
                end

                bitstream_ref(bi:bi+length(bits)-1) = bits;
                bi = bi+length(bits);
            end
        end

        try
            r.reconstruction_demand = mean(reconstruction_map);
            if options.verbose == true
                fprintf('Reconstruction demand: %.4f\n', mean(reconstruction_map));
                fprintf('Need to decode %d symbols having %d symbols\n', ceil(mean(reconstruction_map)*blocks_no), H*W/BLOCK/BLOCK-sum(sum(error_map)));
            end
            
            % Decode the necessary reference symbols
            if strcmp(method, 'general-erasure')
                if options.verbose == true
                    fprintf('Using general erasure model for reference stream decoding.\n');
                end
                if options.use_mex == true
                    bits_decoder = c_rlf_decode(bits_recovered, S_LENGTH, blocks_no, 1-error_map, KEY);
                else
                    bits_decoder = RLF_decode(bits_recovered, S_LENGTH, blocks_no, error_map, KEY);
                end
            elseif strcmp(method, 'self-recovery')
                if options.verbose == true
                    fprintf('Using self-recovery model for reference stream decoding.\n');
                end
                if options.use_mex == true
                    bits_decoder = c_rlf_self_decode(bits_recovered, bitstream_ref, S_LENGTH, 1-error_map, 1-reconstruction_map, KEY);
                else
                    bits_decoder = RLF_self_decode(bits_recovered, S_LENGTH, blocks_no, error_map, KEY, bitstream_ref, reconstruction_map);
                end
            else
                fprintf('WARNING: Decoding method not supported - using self-recovery\n');
                if options.use_mex == true
                    bits_decoder = c_rlf_self_decode(bits_recovered, bitstream_ref, S_LENGTH, 1-error_map, 1-reconstruction_map, KEY);
                else
                    bits_decoder = RLF_self_decode(bits_recovered, S_LENGTH, blocks_no, error_map, KEY, bitstream_ref, reconstruction_map);
                end
            end
            
        catch e
            fprintf('ERROR %s\n', e.message);
            ERR = length(bitstream_ref);
        end
         
        if options.verbose == true
            fprintf('Reference bit-stream errors: %d\n', ERR);
        end
        
        Q_map = [0 1 2 3 4 5 6 7;
                 1 2 3 4 5 6 7 8;
                 2 3 4 5 6 7 8 9;
                 3 4 5 6 7 8 9 10;
                 4 5 6 7 8 9 10 11;
                 5 6 7 8 9 10 11 12;
                 6 7 8 9 10 11 12 13;
                 7 8 9 10 11 12 13 14];
        
        % Reconstruct the Content
        if ERR == 0
            
            imgt_a = double(floor(double(imgt)/dr_factor));
            S_r = blkproc(imgt_a, BLOCK*[1, 1], f);

            data_pointer = 1;
            Ps = sum(sum(P));
            
            for bx = 1:W/BLOCK
                
                for by = 1:H/BLOCK
                                        
                    d = by + (bx-1)*H/BLOCK;

                    if error_map(d) == true
                        data = bits_decoder(data_pointer:data_pointer + Ps - 1);

                        % reconstruct the low quality version of the 
                        % image block content

                        data_index = 1;
                        B = zeros(size(B));

                        Pc = P(indices);
                        Pc(Pc == 0) = [];

                        for p = 1:length(Pc)
                            Q = Pc(p);
                            I = binvec2dec(data(data_index:data_index+Q-1))+1;
                            LM_quanta = LM_quantas{profile_id};
                            codebook = LM_quanta{Q_map(indices(p))+1,Q+1};
                            V = codebook(I);
                            B(indices(p)) = V;
                            data_index = data_index + Q;
                        end

                        S_r(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
                    end
                    data_pointer = data_pointer + Ps;
                end
            end
            
            % Add artificial noise to restored fragments, if requested
            if RANDOMIZE_LSBS == true
                img_fin = uint8(round(dr_factor*blkproc(S_r, BLOCK*[1, 1], fi)));
                for bx = 1:W/BLOCK
                    for by = 1:H/BLOCK
                        d = by + (bx-1)*H/BLOCK;
                        if error_map(d) == true
                            B = img_fin(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
                            B = B + uint8(randint(BLOCK, BLOCK, 8));
                            img_fin(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
                        end
                    end
                end
            else        
                % Otherwise, set 100 as LSBs
                img_fin = blkproc(S_r, BLOCK*[1, 1], fi);
                img_fin = uint8(round(dr_factor*img_fin+pow2(-1+log2(dr_factor))*ones(size(imgt))));

                % Otherwise, clear the LSBs for the whole image
%                 img_fin = uint8(round(dr_factor*blkproc(S_r, BLOCK*[1, 1], fi)));
%                 img_fin = dr_factor*floor(img_fin/dr_factor);
%                 img_fin = bitset(img_fin,LSBS,1);
            end
        end
    else
        if sum(sum(error_map)) == 0
            img_fin = imgt;
        else
            ERR = length(bits_recovered);
        end
    end
    
    % Add missing entries to the resulting structure
    r.B = W*H/BLOCK/BLOCK; % number of image blocks
%     r.K = blocks_no; % number of reference symbols
    r.T = sum(sum(error_map));  % the number of tampered blocks
    r.str_errors = ERR; % number of bit-stream errors after decoding
    r.error_map = error_map; % error map
    r.imgt = imgt; % the tampered image
    r.imgt2 = imgt2; % the tampered image with tinted tampering detection
    r.img_fin = img_fin; % final reconstructed image
    if exist('bits_decoder', 'var')
        r.bitstream = bits_decoder;
    end
end