function [img_low_quality, r] = get_reference_high(img, options, QProfiles, profile_id)
%
% [img_low_quality, r] = get_reference_ref(img, options, QProfiles, profile_id)
%
% Returns the reference image. If QD is not specified in the options structure,
% a profile mapping procedure will be performed.
%
% Parameters:
%
%   img			- input image
%
%   options		- options structure, use self_recovery_defaults() for defaults
%
%   QProfiles	- the QProfiles structure, see setup_qprofiles for details
%
%   profile_id	- ID of the profile for use
%
% Output:
%
%   img_low_quality - the reference image
%
%   r               - details of the reference generation procedure, e.g., psnr
%
% Copyright (C) 2011 Paweł Korus <pkorus@agh.edu.pl>
%
% Version: 2011.12.16
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%

    [H, W] = size(img);
    r = struct();
    
    % Read options
    LSBS = options.LSBS;
    dr_factor = pow2(options.LSBS);
    BLOCK = options.BLOCK;
                    
    % PREPARE A LOW-DETAIL VERSION OF THE IMAGE.

    % Limit the dynamic range of the image
    img_low_dr = double(floor(double(img)/dr_factor));
    
    % Initialize quality profiles
    P_profiles = QProfiles.P_profiles;
    LM_quantas = QProfiles.LM_quantas;
            
    if options.verbose == true
        fprintf('Image: %d x %d px -> %d blocks\n', W, H, W*H/BLOCK/BLOCK);
        fprintf('Using %d LSBs -> %d bits of payload (%d b / block)\n', LSBS, W*H*LSBS, BLOCK*BLOCK*LSBS);
    end
        
    % Transform the low-dynamic range image into spectral domain
    T = dctmtx(8);
    f = @(x) (T*x*T');
    fi = @(x) T'*(x)*T;

    S_b = blkproc(img_low_dr, BLOCK*[1, 1], f);

    % Get a mapping for the zig-zag order

    indices = zigzag(BLOCK);
	P = P_profiles(:,:,profile_id);

    % GENERATE THE RECONSTRUCTION REFERENCE

    for bx = 1:W/BLOCK
        for by = 1:H/BLOCK
            B = S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
            
            % Quantize the block
            [B, I] = quantize_block(B, P, 1, LM_quantas{profile_id});
                        
            if sum(sum(P)) == 0
                S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = 0;
            else
				S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
			end
        end
    end
    
    img_low_quality = blkproc(S_b, BLOCK*[1, 1], fi);
    img_low_quality = uint8(round(dr_factor*img_low_quality+pow2(-1+log2(dr_factor))*ones(size(img))));
    r.img_low_quality = img_low_quality;
    
    % Calculate reconstruction reference quality
    r.ref_psnr = calc_psnr(img, img_low_quality);
    
    if options.verbose == true
        fprintf('Reference quality: %.2f dB\n', r.ref_psnr);
    end