function [QR, QRf] = map_profiles(img, QProfiles, allowed_profiles)
%
% [QR, QRf] = map_profiles(img, QProfiles)
%
% Performs profile mapping - selects best profiles for individual image blocks.
%
% Parameters:
%
%   - img		 - the input image
%
%   - QProfiles  - the QProfiles structure, see setup_qprofiles for details
%
% Output:
% 
%   - QR		 - quality descriptor which maximizes rec. quality for the given
%                  set of quality profiles,
%
%   - QRf		 - full distortion map for the image, i.e., the MSE for 
%                  (bx, by, profile_id)
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%


    block = 8;
    
    T = dctmtx(8);
    f = @(x) (T*x*T');
    fi = @(x) T'*(x)*T;
    
    img = double(img);
    S = blkproc(img, [8,8], f);
    [H, W] = size(img);
    QR = zeros(H/block, W/block);
    QRf = zeros(H/block, W/block, size(QProfiles.P_profiles,3));
    
    for bx = 1:W/block
        for by = 1:H/block
            B = S(1+(by-1)*block:(by)*block, 1+(bx-1)*block:(bx)*block);
            I = img(1+(by-1)*block:(by)*block, 1+(bx-1)*block:(bx)*block);

            dists = zeros(1,size(QProfiles.P_profiles,3));
            
            for i = 1:length(dists)
				if nargin < 3 || length(find(allowed_profiles == i)) > 0
					Br = quantize_block(B, QProfiles.P_profiles(:,:,i), 1, QProfiles.LM_quantas{i});
					dists(i) = mean2(abs(I - fi(Br)));
					QRf(by, bx, i) = dists(i);
				else
					dists(i) = Inf;
				end
            end
            
            dists = round(1000*dists)/1000;
            md = min(dists);
            i = find(dists == md, 1, 'first');
            QR(by, bx) = i;
        end
    end
end