function [gamma_min, gamma_ave, counts, x, y] = plot_demand(weights, lambdas, N, mode)
% 
% [gamma_min, gamma_ave, counts] = plot_demand(weights, lambdas, N, mode)
%
% Plots and calculates the pessimistic tampering rate and the expected
% average maximum tampering rate for a random tampering pattern.
%
% Use mode = 1 for the prioritized tampering pattern and mode = 3 for the oblivious
% random tampering pattern.
%
% Input: 
%
%   - weights  - weights (frequencies) of image blocks with particular
%                lambda values,
%
%   - lambdas  - the corresponding lambda values,
%
%   - N        - number of image blocks
%                CAUTION: Avoid using small block counts (unless you 
%                         know what you're doing) - may mess up
%                         intersections
%
%   - mode:
%
%         Sign controls whether the plot is actually drawn (+ plots).
%
%         The absolute value:
%
%          0 or 1 - default, no changes to the data
%          2 - convert data to single block chunks
%          3 - convert data to single block chunks, then randomize
%
% Output:
%
%   - gamma_min
%
%   - gamma_ave
%
%   - counts
%

    if nargin < 3
        N = 1024;
		mode = 1;
    end
    
    weights = weights/sum(weights);
    counts = round(weights*N);
    counts(end) = counts(end) - (sum(counts) - N);
    
    if abs(mode) >= 2
        
        weights = ones(1,N);
        weights = weights/sum(weights);
        lambdas_temp = ones(1,N);
        li = 1;
        
        for i = 1:length(lambdas)
            lambdas_temp(li:li+counts(i)-1) = lambdas(i);
            li = li + counts(i);
        end
        
        lambdas = lambdas_temp;
        counts = ones(1,N);
        
        if abs(mode) == 3
            lambdas = lambdas(randperm(length(lambdas)));
        end
    end    

    % Sanity check
    if sum(counts) ~= N, fprintf('Insane! Missing %d.\n', N - sum(counts)), return, end;

    % Plot right side progress
    x = linspace(0,1,N+1);   % need N+1 points to represent N ranges
    ry = (1-x)/(sum(weights.*lambdas));

    % Plot left side progress
    y = zeros(1,N+1);
    tgnt = 1./lambdas;

    li = 1;
    A = 0;
    B0 = 0;
    B = B0;
    
    for si = 1:length(counts)

        if ~all(lambdas >= 1) && abs(mode) == 3
            %
            % In the non-aligned mode, the elements with lambdas > 1 are not 
            % grouped in (1\lambda) consecutive blocks - as a result, even
            % lambda >= 1 blocks may become misaligned and the portions cannot
            % be treated as independent.
            %
            % As a result, the tangent raises and the curves are no longer
            % evaluated on a stretched [0,1] range.
            %
            z = lambdas(si)/sum(lambdas.*weights);
            tx = linspace(B,x(li+counts(si)-1),counts(si));
            y(li:li+counts(si)-1) = z*((1 - (1-tx).^ceil(tgnt(si))));

            % Ensure continuous function
            if li > 1
                dx = x(li-1);
                dy = z*((1 - (1-dx).^ceil(tgnt(si))));
                y(li:li+counts(si)) = y(li:li+counts(si)-1) - (dy - y(li-1)); 
            end

        else
            %
            % If all of the blocks are well organized and the reference and 
            % transmission symbols are aligned, the quality levels can be
            % treated independently.
            %
            z = lambdas(si)/sum(lambdas.*weights);
            tx = linspace(B0,1,counts(si)+1);
            if li > 1
                tx = tx(2:end);
            end
            y(li:li+counts(si)-(li>1)) = A + z*(x(li+counts(si)-(li>1))-B)*((1 - (1-tx).^ceil(tgnt(si))));

        end

        if si < length(counts)
            A = y(li+counts(si)-(li>1));
            B = x(li+counts(si)-(li>1));
        end

        li = li+counts(si)+(li==1);
    end

    y(y > 1) = 1;

    kpa = weights.*lambdas./sum(weights.*lambdas);
    kpa = kpa./sum(kpa);
    ty = zeros(size(x));
    for i = 1:length(weights)
        if lambdas(i) >= 1
            ty = ty + kpa(i)*x;
        else
            ty = ty + kpa(i)*(1-(1-x).^(1/lambdas(i)));
        end
    end
    
    % Use linear interpolation - for better accuracy
    xd = 0:0.0001:1;
    yd = interp1(x, y, xd);
    ryd = interp1(x, ry, xd);
    tyd = interp1(x, ty, xd);
    
    [~, xi] = min(abs(yd-ryd));
    [~, xia] = min(abs(ryd-tyd));
    gamma_min = xd(xi);
    gamma_ave = xd(xia);

    % Plotting
    if mode >= 0
        plot(x, ty, 'k:'); 
        axis square; hold on;
        plot(x, ry, 'b');
        plot(x,y, 'r-');
        plot(xd(xi)*[1 1], [0 min(yd(xi), ryd(xi))], 'k--');
        plot(xd(xia)*[1 1], [0 min(yd(xia), ryd(xia))], 'k:');
        ylim([0 1]);
        hold off;
        xlabel('1-\gamma');
        ylabel('Relative symbols count');
        title(sprintf('Estimated \\gamma = %.4f (%.4f)', xd(xi), xd(xia)));
    end
end