function [QD, gamma_out] = qd_optimize(QD, QDf, IMap, gamma, mode, guard_levels, steps, full_lambdas)
% 
% [QD, gamma_out] = qd_optimize(QD, QDf, IMap, desired_gamma, mode, guard_levels, steps, full_lambdas)
%
% Optimization of the quality descriptor to meet the requirements towards
% the tampering rate. 
%
% Params:
%   QD              - the initial quality descriptor
%   QDf             - the profile distortion mapping
%   IMap            - importance map, specifies fragment excluded from 
%                     optimization
%   gamma           - the desired tampering rate bound
%   mode            - optimization mode: 
%                       0 for pessimistic bound,
%                       1 for average.
%   guard_levels    - guard levels, below which quality cannot be decreased
%   steps           - number of blocks that are affected in each iteration
%   full_lambdas    - complete profile-lambda mapping
%
% Output:
%
%   QD              - the optimized quality descriptor
%   gamma_out       - the obtained tampering rate
%
% Copyright (c) 2012 Paweł Korus <pkorus@agh.edu.pl>
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%


% Get the number of available quality profiles
L = size(QDf,3);

% Create quality descrease matrix
R = Inf*ones(size(QD));

for i = 2:L
    Di = QDf(:,:,i);
    Di1 = QDf(:,:,i-1);
    Dd = Di1 - Di;
    for g = guard_levels
        Dd(QD == g) = Inf;
    end
    R(QD == i) = Dd(QD == i);
end

% Disable optimization for selected fragments
R(IMap == 1) = Inf;

[w,l] = qd2weights(QD, full_lambdas);
ri = 1;
gammas = zeros(2,1000);
psnrs = zeros(1,1000);

[gammas(1,ri), gammas(2,ri)] = plot_demand(w(end:-1:1), l(end:-1:1), numel(QD), -1);
psnrs(ri) = 1;

can_optimize = 1;

while can_optimize && gammas(mode+1,ri) < gamma
    
    for s = 1:steps
        % Find element with smallest possible quality loss
        [~, ind] = min(R(:));
        if isinf(Dd(ind))
            can_optimize = 0;
            break;
        end

        % Correct it in the descriptor
        QD(ind) = QD(ind) - 1;

        % Correct the difference
        if QD(ind) - 1 > 0
            Di = QDf(:,:,QD(ind));
            Di1 = QDf(:,:,QD(ind)-1);
            Dd = Di1 - Di;
        end
        for g = guard_levels
            Dd(QD == g) = Inf;
        end
        R(ind) = Dd(ind);
    end
    
    ri = ri + 1;
    psnrs(ri) = ri;
    
    % uncomment to display intermediate results
%     subplot(2,2,1);
%     imagesc(QD); colormap gray;
%     title(sprintf('%d ', unique(QD(:))));
%     [w,l] = qd2weights(QD, full_lambdas);
%     subplot(2,2,2)
%     [gammas(1,ri), gammas(2,ri)] = plot_demand(w(end:-1:1), l(end:-1:1), numel(QD), 1);
%     subplot(3,1,3)
%     plot(psnrs(1:ri), gammas(1,1:ri), psnrs(1:ri), gammas(2,1:ri), [min(psnrs(1:ri)) max(psnrs(1:ri))], gamma*[1 1], 'k:');
% %     xlabel('Iterations');
% %     ylabel('Tampering rates');
% %     legend('Pessimistic', 'Average', 'Target');
%     pause(0.005);

    [w,l] = qd2weights(QD, full_lambdas);
    [gammas(1,ri), gammas(2,ri)] = plot_demand(w(end:-1:1), l(end:-1:1), numel(QD), -1);
end

gamma_out = gammas(mode+1,ri);

end