function QProfiles = setup_qprofiles(options, SA_profiles, lambdas, base_symbol_length, guard_levels)
%
% QProfiles = setup_qprofiles(options, SA_profiles, lambdas, base_symbol_length, guard_levels)
% 
% Setup quality profiles for a given configuration of the self-embedding
% procedure. The following information will be used from the provided
% options structure:
% 
%   - dataset     - name of the training data-set, 'bows', 'sipi' or 'ucid'
%
%   - dm_quantizer- quantization code-book, 'lloyd', 'uniform' or 'lloyd-pi'
%
%   - BLOCK       - block size, 8
%
%   - LSBS        - least significant bit planes eligible for embedding, 1, 2, 3
%
%   - H_LENGTH    - hash length, 32
%
%   - verbose     - use this, if a report on the allocation results is needed
%
% The remaining params: 
%  - SA_profiles         - vec, indicate block type:
%                          0 - default, covers all blocks
%                          1 - low-texture
%                          2 - medium-texture
%                          3 - high-texture
%  - lambdas             - vec, successive quality levels
%  - base_symbol_length  - scalar, symbol length for lambda = 1
%  - guard_levels        - vec, list of guard profiles
%
% For example usage consult setup_qprofiles_main.m
% For more information about the params consult the dissertation.
% 
% The resulting QProfiles structure has the following fields:
% 
%    P_profiles   - allocation matrices for the successive profiles;
%                   a 8 x 8 x P matrix, use P_profiles(:,:,p) to access the 
%                   allocation matrix for the p-th profiles;
%                   This is the main result of the setup procedure.
%
%    LM_quantas   - The appropriate code-books for all of the DCT coefficients;
%                   The code-books are loaded from disk, and chosen according to
%                   allocation results;
%                   Cell array {1,P} with {15,9} cell arrays inside.
%    
%    m_files      - A list of loaded distortion-model files; cell array {1,P}
%    
%    cb_files     - A list of loaded code-book files; cell array {1,P}
%    
%    lambdas      - A vector of reference rates for successive profiles;
%    
%    allocations  - A vector of reference payloads for successive profiles;
%    
%    sa_levels    - A vector of used spatial activity modes:
%                    0 - global mode
%                    1 - low-texture blocks
%                    2 - medium texture blocks
%                    3 - heavy-texture blocks
%    
%    guard_levels - A vector of guard quality profiles; See descriptor design
%                   for more information.
%
%
% Copyright (c) 2012 Paweł Korus <pkorus@agh.edu.pl>
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%

    options.dr_factor = pow2(options.LSBS);
    LB_profiles = lambdas*base_symbol_length;
    
    % if only one argument is given, design with default settings
    if nargin == 1
        SA_profiles = [0   0 0 0 0];
        lambdas = [0.5 1 2 3 4];
        LB_profiles = lambdas*96;
        guard_levels = [2];    
    end
    
    N = length(LB_profiles);
    P_profiles = zeros(options.BLOCK,options.BLOCK,N); 
        
    LM_quantas = cell(1,N);
    m_files = cell(1,N);
    cb_files = cell(1,N);
    
    for i = 1:N
        if LB_profiles(i) > 0
            m_file = sprintf('../quantizers/DM_%d_%s_%d_%s.mat', SA_profiles(i), options.dm_dataset, options.dr_factor, options.dm_quantizer);
            [P, ~, d_sim] = simple_alloc(LB_profiles(i), 'exp', m_file);
            P_profiles(:,:,i) = P;
            cb_file = sprintf('../quantizers/CB_%d_%s_%d_%s.mat', SA_profiles(i), options.dm_dataset, options.dr_factor, options.dm_quantizer);
            load(cb_file);
            LM_quantas{i} = LM_quanta;
            m_files{i} = m_file;
            cb_files{i} = cb_file;
        end
    end
    
    QProfiles = struct();
    QProfiles.P_profiles = P_profiles;
    QProfiles.LM_quantas = LM_quantas;
    QProfiles.lambdas = lambdas;
    QProfiles.m_files = m_files;
    QProfiles.cb_files = cb_files;
    QProfiles.allocations = LB_profiles;
    QProfiles.sa_levels = SA_profiles;
    QProfiles.guard_levels = guard_levels;
    
end