function [bits_out] = RLF_decode(bits_in, block, blocks, error_map, key)
%
% bits_out = RLF_decode(bits_in, block, blocks, error_map, key)
%
% Implementation of the random linear fountain decoder. Uses Gaussian
% elimination for solving the problem.
% 
% Input Parameters:
%
%   - bits_in - input bitstream, vector of 0 and 1
%
%   - block   - symbol length
%
%   - blocks  - the number of original message symbols
%  
%   - error_map - the tampering map, vector of 0 and 1, tampered as 1-s
%
%   - key - integer, the seed for the PRNG
%
% Output:
% 
% bits_out - decoded bitstream, vector of 0 and 1
%
% ------------------------------------------------------------------------
%
% Copyright (C) 2011 Paweł Korus <pkorus@agh.edu.pl>
%
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%
%

    blk = length(bits_in)/block;
    D = zeros(1, blk);
    G = zeros(blk, blocks, 'uint16');
    
    rand('twister', key);
    
    for d = 1:blk
        ro = randint(1, blocks, 2); 
        G(d,:) = ro;
        if nargin > 3
            if error_map(d) == true
                G(d,:) = 0;
            end
        end
    end
    
    % Append symbols to G
    D = zeros(length(bits_in)/block, blocks+block);
    for bi=1:length(bits_in)/block
        D(bi,:) = [G(bi,:) bits_in(1+(bi-1)*block:bi*block)];
    end
    
    G = D;
    
    % Step I
    for i = 1:blocks
        % find rows with ones on i-th position
        rows = i-1+find(G(i:end,i));
        if isempty(rows)
            throw( MException('rlc:insufficientData', 'Not enought input symbols'));
        end
        % select the one with the least ones
        sums = sum(G');
        relevant_sums = sums(rows);
        [x, row_ind] = min(relevant_sums);
        row_ind = rows(row_ind);
        tr = G(row_ind,:);
        % remove the dependency on row_ind form other rows
        for b = rows'
            if b ~= row_ind
                G(b,:) = mod(G(b,:) + tr,2);
            end
        end
        % move the row as the first one
        G(row_ind,:) = G(i,:);
        G(i,:) = tr;
    end
    
    % Step II - back-substitution
    for i = 1:blocks
        % find rows with ones on i-th position
        rows = find(G(1:blocks-i+1,blocks-i+1));
        row_ind = rows(end);
        tr = G(row_ind,:);
        % remove the dependency on row_ind form other rows
        for b = rows'
            if b ~= row_ind
                G(b,:) = mod(G(b,:) + tr,2);
            end
        end
    end
    
    bits_out = G(1:blocks,blocks+1:end);
    bits_out = bits_out';
    bits_out = bits_out(:)';
end